"""
Class for reading data from a 3-brain Biocam system.

See:
https://www.3brain.com/products/single-well/biocam-x

Authors: Alessio Buccino, Robert Wolff
"""

import json
import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _signal_buffer_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

import numpy as np
import json
import warnings
from neo.core import NeoReadWriteError


class BiocamRawIO(BaseRawIO):
    """
    Class for reading data from a Biocam h5 file.

    Parameters
    ----------
    filename: str, default: ''
        The *.h5 file to be read
    fill_gaps_strategy: "zeros" | "synthetic_noise" | None, default: None
        The strategy to fill the gaps in the data when using event-based
        compression. If None and the file is event-based compressed,
        you need to specify a fill gaps strategy:

        * "zeros": the gaps are filled with unsigned 0s (2048). This value is the "0" of the unsigned 12 bits
                   representation of the data.
        * "synthetic_noise": the gaps are filled with synthetic noise.

    Examples
    --------
        >>> import neo.rawio
        >>> r = neo.rawio.BiocamRawIO(filename='biocam.h5')
        >>> r.parse_header()
        >>> print(r)
        >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                                 seg_index=0,
                                                 i_start=0,
                                                 i_stop=1024,
                                                 channel_names=channel_names)
        >>> float_chunk = r.rescale_signal_raw_to_float(raw_chunk,
                                                        dtype='float64',
                                                        channel_indexes=[0, 3, 6])
    """

    extensions = ["h5", "brw"]
    rawmode = "one-file"

    def __init__(self, filename="", fill_gaps_strategy="zeros"):
        BaseRawIO.__init__(self)
        self.filename = filename
        self._fill_gaps_strategy = fill_gaps_strategy

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        self._header_dict = open_biocam_file_header(self.filename)
        self._num_channels = self._header_dict["num_channels"]
        self._num_frames = self._header_dict["num_frames"]
        self._sampling_rate = self._header_dict["sampling_rate"]
        self._filehandle = self._header_dict["file_handle"]
        self._read_function = self._header_dict["read_function"]
        self._channels = self._header_dict["channels"]
        gain = self._header_dict["gain"]
        offset = self._header_dict["offset"]

        # buffer concept cannot be used in this reader because of too complicated dtype across version
        signal_buffers = np.array([], dtype=_signal_stream_dtype)
        signal_streams = np.array([("Signals", "0", "")], dtype=_signal_stream_dtype)

        sig_channels = []
        for c, chan in enumerate(self._channels):
            ch_name = f"ch{chan[0]}-{chan[1]}"
            chan_id = str(c + 1)
            sr = self._sampling_rate  # Hz
            dtype = "uint16"
            units = "uV"
            gain = gain
            offset = offset
            stream_id = "0"
            buffer_id = ""
            sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id, buffer_id))
        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_buffers"] = signal_buffers
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        all_starts = [[0.0]]
        return all_starts[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._num_frames / self._sampling_rate
        all_stops = [[t_stop]]
        return all_stops[block_index][seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._num_frames

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index must be 0")
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._num_frames

        # read functions are different based on the version of biocam
        if self._read_function is readHDF5t_brw4_sparse:
            if self._fill_gaps_strategy is None:
                raise ValueError("Please set `fill_gaps_strategy` to 'zeros' or 'synthetic_noise'.")
            if self._fill_gaps_strategy == "synthetic_noise":
                warnings.warn(
                    "Event-based compression : gaps will be filled with synthetic noise. "
                    "Set `fill_gaps_strategy` to 'zeros' to fill gaps with 0s."
                )
                use_synthetic_noise = True
            elif self._fill_gaps_strategy == "zeros":
                use_synthetic_noise = False
            else:
                raise ValueError("`fill_gaps_strategy` must be 'zeros' or 'synthetic_noise'")

            data = self._read_function(
                self._filehandle, i_start, i_stop, self._num_channels, use_synthetic_noise=use_synthetic_noise
            )
        else:
            data = self._read_function(self._filehandle, i_start, i_stop, self._num_channels)

        # older style data returns array of (n_samples, n_channels), should be a view
        # but if memory issues come up we should doublecheck out how the file is being stored
        if data.ndim > 1:
            if channel_indexes is None:
                channel_indexes = slice(None)
            sig_chunk = data[:, channel_indexes]

        # newer style data returns an initial flat array (n_samples * n_channels)
        # we iterate through channels rather than slicing
        # Due to the fact that Neo and SpikeInterface tend to prefer slices we need to add
        # some careful checks around slicing of None in the case we need to iterate through
        # channels. First check if None. Then check if slice and only if slice check that it is slice(None)
        else:
            if channel_indexes is None:
                channel_indexes = [ch for ch in range(self._num_channels)]
            elif isinstance(channel_indexes, slice):
                start = channel_indexes.start or 0
                stop = channel_indexes.stop or self._num_channels
                step = channel_indexes.step or 1
                channel_indexes = [ch for ch in range(start, stop, step)]

            sig_chunk = np.zeros((i_stop - i_start, len(channel_indexes)), dtype=data.dtype)
            # iterate through channels to prevent loading all channels into memory which can cause
            # memory exhaustion. See https://github.com/SpikeInterface/spikeinterface/issues/3303
            for index, channel_index in enumerate(channel_indexes):
                sig_chunk[:, index] = data[channel_index :: self._num_channels]

        return sig_chunk


def open_biocam_file_header(filename) -> dict:
    """Open a Biocam hdf5 file, read and return the recording info, pick the correct method to access raw data,
    and return this to the caller

    Parameters
    ----------
    filename: str
        The file to be parsed

    Returns
    -------
    dict
        The information necessary to read a biocam file (gain, n_samples, n_channels, etc)."""
    import h5py

    rf = h5py.File(filename, "r")

    if "3BRecInfo" in rf.keys():  # brw v3.x
        # Read recording variables
        rec_vars = rf.require_group("3BRecInfo/3BRecVars/")
        bit_depth = rec_vars["BitDepth"][0]
        max_uv = rec_vars["MaxVolt"][0]
        min_uv = rec_vars["MinVolt"][0]
        num_frames = rec_vars["NRecFrames"][0]
        sampling_rate = rec_vars["SamplingRate"][0]
        signal_inv = rec_vars["SignalInversion"][0]

        # Get the actual number of channels used in the recording
        file_format = rf["3BData"].attrs.get("Version", None)
        format_100 = False
        if file_format == 100:
            num_channels = len(rf["3BData/Raw"][0])
            format_100 = True
        elif file_format in (101, 102) or file_format is None:
            num_channels = int(rf["3BData/Raw"].shape[0] / num_frames)
        else:
            raise NeoReadWriteError("Unknown data file format.")

        # get channels
        channels = rf["3BRecInfo/3BMeaStreams/Raw/Chs"][:]

        # determine correct function to read data
        if format_100:
            if signal_inv == 1:
                read_function = readHDF5t_100
            elif signal_inv == -1:
                read_function = readHDF5t_100_i
            else:
                raise NeoReadWriteError("Unknown signal inversion")
        else:
            if signal_inv == 1:
                read_function = readHDF5t_101
            elif signal_inv == -1:
                read_function = readHDF5t_101_i
            else:
                raise NeoReadWriteError("Unknown signal inversion")

        gain = (max_uv - min_uv) / (2**bit_depth)
        offset = min_uv

        return dict(
            file_handle=rf,
            num_frames=num_frames,
            sampling_rate=sampling_rate,
            num_channels=num_channels,
            channels=channels,
            file_format=file_format,
            signal_inv=signal_inv,
            read_function=read_function,
            gain=gain,
            offset=offset,
        )
    else:  # brw v4.x
        # Read recording variables
        experiment_settings = json.JSONDecoder().decode(rf["ExperimentSettings"][0].decode())
        max_uv = experiment_settings["ValueConverter"]["MaxAnalogValue"]
        min_uv = experiment_settings["ValueConverter"]["MinAnalogValue"]
        max_digital = experiment_settings["ValueConverter"]["MaxDigitalValue"]
        min_digital = experiment_settings["ValueConverter"]["MinDigitalValue"]
        scale_factor = experiment_settings["ValueConverter"]["ScaleFactor"]
        sampling_rate = experiment_settings["TimeConverter"]["FrameRate"]
        num_frames = rf["TOC"][-1, -1]

        num_channels = None
        well_ID = None
        for well_ID in rf:
            if well_ID.startswith("Well_"):
                num_channels = len(rf[well_ID]["StoredChIdxs"])
                if "Raw" in rf[well_ID]:
                    if len(rf[well_ID]["Raw"]) % num_channels:
                        raise NeoReadWriteError(
                            f"Length of raw data array is not multiple of channel number in {well_ID}"
                        )
                    num_frames = len(rf[well_ID]["Raw"]) // num_channels
                    break
                elif "EventsBasedSparseRaw" in rf[well_ID]:
                    # Not sure how to check for this with sparse data
                    pass

        if num_channels is not None:
            num_channels_x = num_channels_y = int(np.sqrt(num_channels))
        else:
            raise NeoReadWriteError("No Well found in the file")

        if num_channels_x * num_channels_y != num_channels:
            raise NeoReadWriteError(f"Cannot determine structure of the MEA plate with {num_channels} channels")
        channels = 1 + np.concatenate(np.transpose(np.meshgrid(range(num_channels_x), range(num_channels_y))))

        gain = scale_factor * (max_uv - min_uv) / (max_digital - min_digital)
        offset = min_uv
        if "Raw" in rf[well_ID]:
            read_function = readHDF5t_brw4
        elif "EventsBasedSparseRaw" in rf[well_ID]:
            read_function = readHDF5t_brw4_sparse

        return dict(
            file_handle=rf,
            num_frames=num_frames,
            sampling_rate=sampling_rate,
            num_channels=num_channels,
            channels=channels,
            read_function=read_function,
            gain=gain,
            offset=offset,
        )


######################################################################
# Helper functions to obtain the raw data split by Biocam version.


# return the full array for the old datasets
def readHDF5t_100(rf, t0, t1, nch):
    return rf["3BData/Raw"][t0:t1]


def readHDF5t_100_i(rf, t0, t1, nch):
    return 4096 - rf["3BData/Raw"][t0:t1]


# return flat array that we will iterate through
def readHDF5t_101(rf, t0, t1, nch):
    return rf["3BData/Raw"][nch * t0 : nch * t1]


def readHDF5t_101_i(rf, t0, t1, nch):
    return 4096 - rf["3BData/Raw"][nch * t0 : nch * t1]


def readHDF5t_brw4(rf, t0, t1, nch):
    for key in rf:
        if key.startswith("Well_"):
            return rf[key]["Raw"][nch * t0 : nch * t1]


def readHDF5t_brw4_sparse(rf, t0, t1, nch, use_synthetic_noise=False):

    # noise_std = None
    start_frame = t0
    num_frames = t1 - t0
    for well_ID in rf:
        if well_ID.startswith("Well_"):
            break
    # initialize an empty (fill with zeros) data collection
    data = np.zeros((nch, num_frames), dtype=np.uint16)
    if not use_synthetic_noise:
        # Will read as 0s after 12 bits signed conversion
        data.fill(2048)
    else:
        # fill the data collection with Gaussian noise if requested
        data = generate_synthetic_noise(rf, data, well_ID, start_frame, num_frames)  # , std=noise_std)
    # fill the data collection with the decoded event based sparse raw data
    data = decode_event_based_raw_data(rf, data, well_ID, start_frame, num_frames)

    return data.T


def decode_event_based_raw_data(rf, data, well_ID, start_frame, num_frames):
    # Source: Documentation by 3Brain
    # https://gin.g-node.org/NeuralEnsemble/ephy_testing_data/src/master/biocam/documentation_brw_4.x_bxr_3.x_bcmp_1.x_in_brainwave_5.x_v1.1.3.pdf
    # collect the TOCs
    toc = np.array(rf["TOC"])
    events_toc = np.array(rf[well_ID]["EventsBasedSparseRawTOC"])
    # from the given start position and duration in frames, localize the corresponding event positions
    # using the TOC
    toc_start_idx = np.searchsorted(toc[:, 1], start_frame)
    toc_end_idx = min(np.searchsorted(toc[:, 1], start_frame + num_frames, side="right") + 1, len(toc) - 1)
    events_start_pos = events_toc[toc_start_idx]
    events_end_pos = events_toc[toc_end_idx]
    # decode all data for the given well ID and time interval
    binary_data = rf[well_ID]["EventsBasedSparseRaw"][events_start_pos:events_end_pos]
    binary_data_length = len(binary_data)
    pos = 0
    while pos < binary_data_length:
        ch_idx = int.from_bytes(binary_data[pos : pos + 4], byteorder="little")
        pos += 4
        ch_data_length = int.from_bytes(binary_data[pos : pos + 4], byteorder="little")
        pos += 4
        ch_data_pos = pos
        while pos < ch_data_pos + ch_data_length:
            from_inclusive = int.from_bytes(binary_data[pos : pos + 8], byteorder="little")
            pos += 8
            to_exclusive = int.from_bytes(binary_data[pos : pos + 8], byteorder="little")
            pos += 8
            range_data_pos = pos
            for j in range(from_inclusive, to_exclusive):
                if j >= start_frame + num_frames:
                    break
                if j >= start_frame:
                    data[ch_idx][j - start_frame] = int.from_bytes(
                        binary_data[range_data_pos : range_data_pos + 2], byteorder="little"
                    )
                range_data_pos += 2
            pos += (to_exclusive - from_inclusive) * 2

    return data


def generate_synthetic_noise(rf, data, well_ID, start_frame, num_frames):
    # Source: Documentation by 3Brain
    # https://gin.g-node.org/NeuralEnsemble/ephy_testing_data/src/master/biocam/documentation_brw_4.x_bxr_3.x_bcmp_1.x_in_brainwave_5.x_v1.1.3.pdf
    # collect the TOCs
    toc = np.array(rf["TOC"])
    noise_toc = np.array(rf[well_ID]["NoiseTOC"])
    # from the given start position in frames, localize the corresponding noise positions
    # using the TOC
    toc_start_idx = np.searchsorted(toc[:, 1], start_frame)
    noise_start_pos = noise_toc[toc_start_idx]
    noise_end_pos = noise_start_pos
    for i in range(toc_start_idx + 1, len(noise_toc)):
        next_pos = noise_toc[i]
        if next_pos > noise_start_pos:
            noise_end_pos = next_pos
        break
    if noise_end_pos == noise_start_pos:
        for i in range(toc_start_idx - 1, 0, -1):
            previous_pos = noise_toc[i]
            if previous_pos < noise_start_pos:
                noise_end_pos = noise_start_pos
                noise_start_pos = previous_pos
                break
    # obtain the noise info at the start position
    noise_ch_idx = rf[well_ID]["NoiseChIdxs"][noise_start_pos:noise_end_pos]
    noise_mean = rf[well_ID]["NoiseMean"][noise_start_pos:noise_end_pos]
    noise_std = rf[well_ID]["NoiseStdDev"][noise_start_pos:noise_end_pos]

    noise_length = noise_end_pos - noise_start_pos
    noise_info = {}
    mean_collection = []
    std_collection = []
    for i in range(1, noise_length):
        noise_info[noise_ch_idx[i]] = [noise_mean[i], noise_std[i]]
        mean_collection.append(noise_mean[i])
        std_collection.append(noise_std[i])
    # calculate the median mean and standard deviation of all channels to be used for
    # invalid channels
    median_mean = np.median(mean_collection)
    median_std = np.median(std_collection)
    # fill with Gaussian noise
    for ch_idx in range(len(data)):
        if ch_idx in noise_info:
            data[ch_idx] = np.array(
                np.random.normal(noise_info[ch_idx][0], noise_info[ch_idx][1], num_frames), dtype=np.uint16
            )
        else:
            data[ch_idx] = np.array(np.random.normal(median_mean, median_std, num_frames), dtype=np.uint16)

    return data



"""
baseio
======

Classes
-------

BaseIO        - abstract class which should be overridden, managing how a
                file will load/write its data

If you want a model for developing a new IO start from exampleIO.
"""

from __future__ import annotations
from pathlib import Path
from collections.abc import Sequence
import logging

from neo import logging_handler
from neo.core import (
    AnalogSignal,
    Block,
    Epoch,
    Event,
    Group,
    IrregularlySampledSignal,
    ChannelView,
    Segment,
    SpikeTrain,
    ImageSequence,
    RectangularRegionOfInterest,
    CircularRegionOfInterest,
    PolygonRegionOfInterest,
    NeoReadWriteError,
)

read_error = "This type is not supported by this file format for reading"
write_error = "This type is not supported by this file format for writing"


class BaseIO:
    """
    Generic class to handle all the file read/write methods for the key objects
    of the core class. This template is file-reading/writing oriented but it
    can also handle data read from/written to a database such as TDT sytem
    tanks or SQLite files.

    This is an abstract class that will be subclassed for each format
    The key methods of the class are:
        - ``read()`` - Read the whole object structure, return a list of Block
                objects
        - ``read_block(lazy=True, **params)`` - Read Block object
                from file with some parameters
        - ``read_segment(lazy=True, **params)`` - Read Segment
                object from file with some parameters
        - ``read_spiketrainlist(lazy=True, **params)`` - Read
                SpikeTrainList object from file with some parameters
        - ``write()`` - Write the whole object structure
        - ``write_block(**params)``    - Write Block object to file with some
                parameters
        - ``write_segment(**params)``    - Write Segment object to file with
                some parameters
        - ``write_spiketrainlist(**params)`` - Write SpikeTrainList object to
                file with some parameters

    The class can also implement these methods:
        - ``read_XXX(lazy=True, **params)``
        - ``write_XXX(**params)``
        where XXX could be one of the objects supported by the IO

    Each class is able to declare what can be accessed or written directly
    discribed by **readable_objects** and **readable_objects**.
    The object types can be one of the classes defined in neo.core
    (Block, Segment, AnalogSignal, ...)

    Each class does not necessary support all the whole neo hierarchy but part
    of it.
    This is described with **supported_objects**.
    All IOs must support at least Block with a read_block()


    ** start a new IO **
    If you want to implement your own file format, you should create a class
    that will inherit from this BaseFile class and implement the previous
    methods.
    See ExampleIO in exampleio.py
    """

    is_readable = False
    is_writable = False

    supported_objects = []
    readable_objects = []
    writeable_objects = []

    support_lazy = False

    read_params = {}
    write_params = {}

    name = "BaseIO"
    description = ""
    extensions = []

    mode = "file"  # or 'fake' or 'dir' or 'database'

    def __init__(self, filename: str | Path = None, **kargs):
        self.filename = str(filename)
        # create a logger for the IO class
        fullname = self.__class__.__module__ + "." + self.__class__.__name__
        self.logger = logging.getLogger(fullname)
        # create a logger for 'neo' and add a handler to it if it doesn't
        # have one already.
        # (it will also not add one if the root logger has a handler)
        corename = self.__class__.__module__.split(".")[0]
        corelogger = logging.getLogger(corename)
        rootlogger = logging.getLogger()
        if not corelogger.handlers and not rootlogger.handlers:
            corelogger.addHandler(logging_handler)

    ######## General read/write methods #######################
    def read(self, lazy: bool = False, **kargs):
        """
        Return all data from the file as a list of Blocks

        Parameters
        ----------
        lazy: bool, default: False
            Whether to lazily load the data (True) or to load into memory (False)
        kargs: dict
            IO specific additional arguments

        Returns
        ------
        block_list: list[neo.core.Block]
            Returns all the data from the file as Blocks
        """
        if lazy and not self.support_lazy:
            raise NeoReadWriteError("This IO module does not support lazy loading")
        if Block in self.readable_objects:
            if hasattr(self, "read_all_blocks") and callable(getattr(self, "read_all_blocks")):
                return self.read_all_blocks(lazy=lazy, **kargs)
            return [self.read_block(lazy=lazy, **kargs)]
        elif Segment in self.readable_objects:
            bl = Block(name="One segment only")
            seg = self.read_segment(lazy=lazy, **kargs)
            bl.segments.append(seg)
            bl.check_relationships()
            return [bl]
        else:
            raise NotImplementedError

    def write(self, bl, **kargs):
        """
        Writes a given block if IO supports writing

        Parameters
        ----------
        bl: neo.core.Block
            The neo Block to be written
        kargs: dict
            IO specific additional arguments

        """
        if Block in self.writeable_objects:
            if isinstance(bl, Sequence):
                if not hasattr(self, "write_all_blocks"):
                    raise NeoReadWriteError(f"{self.__class__.__name__} does not offer to store a sequence of blocks")
                self.write_all_blocks(bl, **kargs)
            else:
                self.write_block(bl, **kargs)
        elif Segment in self.writeable_objects:
            if len(bl.segments) != 1:
                raise NeoReadWriteError(
                    f"{self.__class__.__name__} is based on segment so if you try to write a block it "
                    + "must contain only one Segment"
                )
            self.write_segment(bl.segments[0], **kargs)
        else:
            raise NotImplementedError

    ######## All individual read methods #######################
    def read_block(self, **kargs):
        if Block not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_segment(self, **kargs):
        if Segment not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_spiketrain(self, **kargs):
        if SpikeTrain not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_analogsignal(self, **kargs):
        if AnalogSignal not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_imagesequence(self, **kargs):
        if ImageSequence not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_rectangularregionofinterest(self, **kargs):
        if RectangularRegionOfInterest not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_circularregionofinterest(self, **kargs):
        if CircularRegionOfInterest not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_polygonregionofinterest(self, **kargs):
        if PolygonRegionOfInterest not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_irregularlysampledsignal(self, **kargs):
        if IrregularlySampledSignal not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_channelview(self, **kargs):
        if ChannelView not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_event(self, **kargs):
        if Event not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_epoch(self, **kargs):
        if Epoch not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    def read_group(self, **kargs):
        if Group not in self.readable_objects:
            raise NeoReadWriteError(read_error)

    ######## All individual write methods #######################
    def write_block(self, bl, **kargs):
        if Block not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_segment(self, seg, **kargs):
        if Segment not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_spiketrain(self, sptr, **kargs):
        if SpikeTrain not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_analogsignal(self, anasig, **kargs):
        if AnalogSignal not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_imagesequence(self, imseq, **kargs):
        if ImageSequence not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_rectangularregionofinterest(self, rectroi, **kargs):
        if RectangularRegionOfInterest not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_circularregionofinterest(self, circroi, **kargs):
        if CircularRegionOfInterest not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_polygonregionofinterest(self, polyroi, **kargs):
        if PolygonRegionOfInterest not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_irregularlysampledsignal(self, irsig, **kargs):
        if IrregularlySampledSignal not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_channelview(self, chv, **kargs):
        if ChannelView not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_event(self, ev, **kargs):
        if Event not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_epoch(self, ep, **kargs):
        if Epoch not in self.writeable_objects:
            raise NeoReadWriteError(write_error)

    def write_group(self, group, **kargs):
        if Group not in self.writeable_objects:
            raise NeoReadWriteError(write_error)



import mmap
import numpy as np


def get_memmap_shape(filename, dtype, num_channels=None, offset=0):
    dtype = np.dtype(dtype)
    with open(filename, mode="rb") as f:
        f.seek(0, 2)
        flen = f.tell()
        bytes = flen - offset
        if bytes % dtype.itemsize != 0:
            raise ValueError("Size of available data is not a multiple of the data-type size.")
        size = bytes // dtype.itemsize
        if num_channels is None:
            shape = (size,)
        else:
            shape = (size // num_channels, num_channels)
        return shape


def get_memmap_chunk_from_opened_file(fid, num_channels, start, stop, dtype, file_offset=0):
    """
    Utility function to get a chunk as a memmap array directly from an opened file.
    Using this instead memmap can avoid memmory consumption when multiprocessing.

    Similar mechanism is used in spikeinterface.

    """
    bytes_per_sample = num_channels * dtype.itemsize

    # Calculate byte offsets
    start_byte = file_offset + start * bytes_per_sample
    end_byte = file_offset + stop * bytes_per_sample

    # Calculate the length of the data chunk to load into memory
    length = end_byte - start_byte

    # The mmap offset must be a multiple of mmap.ALLOCATIONGRANULARITY
    memmap_offset, start_offset = divmod(start_byte, mmap.ALLOCATIONGRANULARITY)
    memmap_offset *= mmap.ALLOCATIONGRANULARITY

    # Adjust the length so it includes the extra data from rounding down
    # the memmap offset to a multiple of ALLOCATIONGRANULARITY
    length += start_offset

    memmap_obj = mmap.mmap(fid.fileno(), length=length, access=mmap.ACCESS_READ, offset=memmap_offset)

    arr = np.ndarray(
        shape=((stop - start), num_channels),
        dtype=dtype,
        buffer=memmap_obj,
        offset=start_offset,
    )

    return arr



""" """

import unittest

from neo.rawio.micromedrawio import MicromedRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

import numpy as np


class TestMicromedRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = MicromedRawIO
    entities_to_download = ["micromed"]
    entities_to_test = [
        "micromed/File_micromed_1.TRC",
        "micromed/File_mircomed2.TRC",
        "micromed/File_mircomed2_2segments.TRC",
    ]

    def test_micromed_multi_segments(self):
        file_full = self.get_local_path("micromed/File_mircomed2.TRC")
        file_splitted = self.get_local_path("micromed/File_mircomed2_2segments.TRC")

        # the second file contains 2 pieces of the first file
        # so it is 2 segments with the same traces but reduced
        # note that traces in the splited can differ at the very end of the cut

        reader1 = MicromedRawIO(file_full)
        reader1.parse_header()
        assert reader1.segment_count(block_index=0) == 1
        assert reader1.get_signal_t_start(block_index=0, seg_index=0, stream_index=0) == 0.0
        traces1 = reader1.get_analogsignal_chunk(stream_index=0)

        reader2 = MicromedRawIO(file_splitted)
        reader2.parse_header()
        print(reader2)
        assert reader2.segment_count(block_index=0) == 2

        # check that pieces of the second file is equal to the first file (except a truncation at the end)
        for seg_index in range(2):
            t_start = reader2.get_signal_t_start(block_index=0, seg_index=seg_index, stream_index=0)
            assert t_start > 0
            sr = reader2.get_signal_sampling_rate(stream_index=0)
            ind_start = int(t_start * sr)
            traces2 = reader2.get_analogsignal_chunk(block_index=0, seg_index=seg_index, stream_index=0)
            traces1_chunk = traces1[ind_start : ind_start + traces2.shape[0]]
            # we remove the last 100 sample because tools that cut traces is truncating the last buffer
            assert np.array_equal(traces2[:-100], traces1_chunk[:-100])


if __name__ == "__main__":
    unittest.main()



"""
NeuroNexus has their own file format based on their Allego Recording System
https://www.neuronexus.com/webinar/allego-software-updates/

The format involves 3 files:
  * The *.xdat.json metadata file
  * The *_data.xdat binary file of all raw data
  * The *_timestamps.xdat binary file of the timestamp data

Based on sample data is appears that the binary file is always a float32 format
Other information can be found within the metadata json file


The metadata file has a pretty complicated structure as far as I can tell
a lot of which is dedicated to probe information, which won't be handle at the
the Neo level.

It appears that the metadata['status'] provides most of the information necessary
for generating the initial memory map (global sampling frequency), n_channels,
n_samples.

metadata['sapiens_base']['biointerface_map'] provides all the channel specific information
like channel_names, channel_ids, channel_types.

An additional note on channels. It appears that analog channels are called `pri` or
`ai0` within the metadata whereas digital channels are called `din0` or `dout0`.
In this first implementation it is up to the user to do the appropriate channel slice
to only get the data they want. This is a buffer-based approach that Sam likes.
Eventually we will try to divide these channels into streams (analog vs digital) or
we can come up with a work around if users open an issue requesting this.

Zach McKenzie

"""

from __future__ import annotations
from pathlib import Path
import json
import datetime
import sys
import re

import numpy as np

from .baserawio import (
    BaseRawWithBufferApiIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _signal_buffer_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


class NeuroNexusRawIO(BaseRawWithBufferApiIO):

    extensions = ["xdat", "json"]
    rawmode = "one-file"

    def __init__(self, filename: str | Path = ""):
        """
        The Allego NeuroNexus reader for the `xdat` file format

        Parameters
        ----------
        filename: str | Path, default: ''
            The filename of the metadata file should end in .xdat.json

        Notes
        -----
        * The format involves 3 files:
            * The *.xdat.json metadata file
            * The *_data.xdat binary file of all raw data
            * The *_timestamps.xdat binary file of the timestamp data

        From the metadata the other two files are located within the same directory
        and loaded.

        * The metadata is stored as the metadata attribute for individuals hoping
        to extract probe information, but Neo itself does not load any of the probe
        information

        Examples
        --------
        >>> from neo.rawio import NeuronexusRawIO
        >>> reader = NeuronexusRawIO(filename='abc.xdat.json')
        >>> reader.parse_header()
        >>> raw_chunk = reader.get_analogsignal_chunk(block_index=0
                seg_index=0,
                stream_index=0)
        # this isn't necessary for this reader since data is stored as float uV, but
        # this is included in case there is a future change to the format
        >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk, stream_index=0)

        """

        BaseRawWithBufferApiIO.__init__(self)

        if not Path(filename).is_file():
            raise FileNotFoundError(f"The metadata file {filename} was not found")
        if Path(filename).suffix != ".json":
            raise NeoReadWriteError(
                f"The json metadata file should be given, filename entered is {Path(filename).stem}"
            )
        meta_filename = Path(filename)
        binary_file = meta_filename.parent / (meta_filename.stem.split(".")[0] + "_data.xdat")

        if not binary_file.exists() and not binary_file.is_file():
            raise FileNotFoundError(f"The data.xdat file {binary_file} was not found. Is it in the same directory?")
        timestamp_file = meta_filename.parent / (meta_filename.stem.split(".")[0] + "_timestamp.xdat")
        if not timestamp_file.exists() and not timestamp_file.is_file():
            raise FileNotFoundError(
                f"The timestamps.xdat file {timestamp_file} was not found. Is it in the same directory?"
            )

        self.filename = filename
        self.binary_file = binary_file
        self.timestamp_file = timestamp_file

    def _source_name(self):
        # return the metadata filename only
        return self.filename

    def _parse_header(self):

        # read metadata
        self.metadata = self.read_metadata(self.filename)

        # Collect information necessary for memory map
        self._sampling_frequency = self.metadata["status"]["samp_freq"]
        self._n_samples, self._n_channels = self.metadata["status"]["shape"]
        # Stored as a simple float32 binary file
        BINARY_DTYPE = "float32"
        binary_file = self.binary_file
        timestamp_file = self.timestamp_file

        # the will cretae a memory map with teh generic mechanism
        buffer_id = "0"
        self._buffer_descriptions = {0: {0: {}}}
        self._buffer_descriptions[0][0][buffer_id] = {
            "type": "raw",
            "file_path": str(binary_file),
            "dtype": BINARY_DTYPE,
            "order": "C",
            "file_offset": 0,
            "shape": (self._n_samples, self._n_channels),
        }
        # Make the memory map for timestamp
        self._timestamps = np.memmap(
            timestamp_file,
            dtype=np.int64,  # this is from the allego sample reader timestamps are np.int64
            mode="r",
            offset=0,  # headerless binary file
        )

        # We can do a quick timestamp check to make sure it is the correct timestamp data for the
        # given metadata
        if self._timestamps[0] != self.metadata["status"]["timestamp_range"][0]:
            metadata_start = self.metadata["status"]["timestamp_range"][0]
            data_start = self._teimstamps[0]
            raise NeoReadWriteError(
                f"The metadata indicates a different starting timestamp {metadata_start} than the data starting timestamp {data_start}"
            )

        # organize the channels
        signal_channels = []
        channel_info = self.metadata["sapiens_base"]["biointerface_map"]

        # as per dicussion with the Neo/SpikeInterface teams stream_id will become buffer_id
        # and because all data is stored in the same buffer stream for the moment all channels
        # will be in stream_id = 0. In the future this will be split into sub_streams based on
        # type but for now it will be the end-users responsability for this.
        stream_id = "0"  # hard-coded see note above
        buffer_id = "0"
        for channel_index, channel_name in enumerate(channel_info["chan_name"]):
            channel_id = channel_info["ntv_chan_name"][channel_index]
            # 'ai0' indicates analog data which is stored as microvolts
            if channel_info["chan_type"][channel_index] == "ai0":
                units = "uV"
            # 'd' means digital. Per discussion with neuroconv users the use of
            # 'a.u.' makes the units clearer
            elif channel_info["chan_type"][channel_index][0] == "d":
                units = "a.u."
            # aux channel
            else:
                units = "V"

            signal_channels.append(
                (
                    channel_name,
                    channel_id,
                    self._sampling_frequency,
                    BINARY_DTYPE,
                    units,
                    1,  # no gain
                    0,  # no offset
                    stream_id,
                    buffer_id,
                )
            )

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        buffer_id = "0"
        signal_buffers = np.array([("", buffer_id)], dtype=_signal_buffer_dtype)

        stream_ids = np.unique(signal_channels["stream_id"])
        signal_streams = np.zeros(stream_ids.size, dtype=_signal_stream_dtype)
        signal_streams["id"] = [str(stream_id) for stream_id in stream_ids]
        # One unique buffer
        signal_streams["buffer_id"] = buffer_id
        self._stream_buffer_slice = {}
        for stream_index, stream_id in enumerate(stream_ids):
            name = stream_id_to_stream_name.get(int(stream_id), "")
            signal_streams["name"][stream_index] = name
            chan_inds = np.flatnonzero(signal_channels["stream_id"] == stream_id)
            self._stream_buffer_slice[stream_id] = chan_inds

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # Put all the necessary info in the header
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_buffers"] = signal_buffers
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Add the minimum annotations
        self._generate_minimal_annotations()

        # date comes out as: '2024-07-01T13:04:49.4972245-04:00' so in ISO format
        datetime_string = self.metadata["status"]["start_time"]

        # Python 3.10 and older expect iso format to only have 3 or 6 decimal places
        if sys.version_info.minor < 11:
            datetime_string = re.sub(r"(\.\d{6})\d+", r"\1", datetime_string)

        rec_datetime = datetime.datetime.fromisoformat(datetime_string)

        bl_annotations = self.raw_annotations["blocks"][0]
        seg_annotations = bl_annotations["segments"][0]
        for d in (bl_annotations, seg_annotations):
            d["rec_datetime"] = rec_datetime

    def _segment_t_stop(self, block_index, seg_index):

        t_stop = self.metadata["status"]["t_range"][1]
        return t_stop

    def _segment_t_start(self, block_index, seg_index):

        t_start = self.metadata["status"]["t_range"][0]
        return t_start

    def _get_signal_t_start(self, block_index, seg_index, stream_index):

        t_start = self.metadata["status"]["t_range"][0]
        return t_start

    #######################################
    # Helper Functions

    def read_metadata(self, fname_metadata):
        """
        Metadata is just a heavily nested json file

        Parameters
        ----------
        fname_metada: str | Path
            The *.xdat.json file for the current recording

        Returns
        -------
        metadata: dict
            Returns the metadata as a dictionary"""

        fname_metadata = Path(fname_metadata)
        with open(fname_metadata, "rb") as read_file:
            metadata = json.load(read_file)

        return metadata

    def _get_analogsignal_buffer_description(self, block_index, seg_index, buffer_id):
        return self._buffer_descriptions[block_index][seg_index][buffer_id]


# this is pretty useless right now, but I think after a
# refactor with sub streams we could adapt this for the sub-streams
# so let's leave this here for now :)
stream_id_to_stream_name = {"0": "Neuronexus Allego Data"}



"""
Experimental module to export a rawio reader that support the buffer_description API
to xarray dataset using zarr specification format v2.

A block/segment/stream correspond to one xarray.DataSet

A xarray.DataTree can also be expose to get all at block/segment/stream

Note :
  * only some IOs support this at the moment has_buffer_description_api()=True
"""

import numpy as np

import base64


def to_zarr_v2_reference(rawio_reader, block_index=0, seg_index=0, buffer_id=None):
    """
    Transform the buffer_description_api into a dict ready for the xarray API 'reference://'


    See https://fsspec.github.io/kerchunk/spec.html
    See https://docs.xarray.dev/en/latest/user-guide/io.html#kerchunk

    See https://zarr-specs.readthedocs.io/en/latest/v2/v2.0.html


    Usefull read also https://github.com/saalfeldlab/n5

    """

    # TODO later implement zarr v3

    # rawio_reader.
    signal_buffers = rawio_reader.header["signal_buffers"]

    buffer_index = list(signal_buffers["id"]).index(buffer_id)

    buffer_name = signal_buffers["name"][buffer_index]

    rfs = dict()
    rfs["version"] = 1
    rfs["refs"] = dict()
    rfs["refs"][".zgroup"] = dict(zarr_format=2)
    zattrs = dict(name=buffer_name)
    rfs["refs"][".zattrs"] = zattrs

    descr = rawio_reader.get_analogsignal_buffer_description(
        block_index=block_index, seg_index=seg_index, buffer_id=buffer_id
    )

    if descr["type"] == "raw":

        # channel : small enough can be internal with base64
        mask = rawio_reader.header["signal_channels"]["buffer_id"] == buffer_id
        channels = rawio_reader.header["signal_channels"][mask]
        channel_ids = channels["id"]
        base64_encoded = base64.b64encode(channel_ids.tobytes())
        rfs["refs"]["channel/0"] = "base64:" + base64_encoded.decode()
        zarray = dict(
            chunks=channel_ids.shape,
            compressor=None,
            dtype=channel_ids.dtype.str,
            fill_value=None,
            filters=None,
            order="C",
            shape=channel_ids.shape,
            zarr_format=2,
        )
        zattrs = dict(
            _ARRAY_DIMENSIONS=["channel"],
        )
        rfs["refs"]["channel/.zattrs"] = zattrs
        rfs["refs"]["channel/.zarray"] = zarray

        # traces buffer
        dtype = np.dtype(descr["dtype"])
        zarray = dict(
            chunks=descr["shape"],
            compressor=None,
            dtype=dtype.str,
            fill_value=None,
            filters=None,
            order=descr["order"],
            shape=descr["shape"],
            zarr_format=2,
        )
        zattrs = dict(
            _ARRAY_DIMENSIONS=["time", "channel"],
            name=buffer_name,
        )
        units = np.unique(channels["units"])
        if units.size == 1:
            zattrs["units"] = units[0]
        gain = np.unique(channels["gain"])
        offset = np.unique(channels["offset"])
        if gain.size == 1 and offset.size:
            zattrs["scale_factor"] = gain[0]
            zattrs["add_offset"] = offset[0]
        zattrs["sampling_rate"] = float(channels["sampling_rate"][0])

        # unique big chunk
        # TODO later : optional split in several small chunks
        array_size = np.prod(descr["shape"], dtype="int64")
        chunk_size = array_size * dtype.itemsize
        rfs["refs"]["traces/0.0"] = [str(descr["file_path"]), descr["file_offset"], chunk_size]
        rfs["refs"]["traces/.zarray"] = zarray
        rfs["refs"]["traces/.zattrs"] = zattrs

    elif descr["type"] == "hdf5":
        raise NotImplementedError
    else:
        raise ValueError(f"buffer_description type not handled {descr['type']}")

    # TODO later channel array_annotations

    return rfs


def to_xarray_dataset(rawio_reader, block_index=0, seg_index=0, buffer_id=None):
    """
    Utils fonction that transorm an instance a rawio into a xarray.Dataset
    with lazy access.
    This works only for rawio class that return True with has_buffer_description_api() and hinerits from
    BaseRawWithBufferApiIO.


    Note : the original idea of the function is from Ben Dichter in this page
    https://gist.github.com/bendichter/30a9afb34b2178098c99f3b01fe72e75
    """
    import xarray as xr

    rfs = to_zarr_v2_reference(rawio_reader, block_index=block_index, seg_index=seg_index, buffer_id=buffer_id)

    ds = xr.open_dataset(
        "reference://",
        mask_and_scale=True,
        engine="zarr",
        backend_kwargs={
            "storage_options": dict(
                fo=rfs,
                remote_protocol="file",
            ),
            "consolidated": False,
        },
    )
    return ds


def to_xarray_datatree(rawio_reader):
    """
    Expose a neo.rawio reader class to a xarray DataTree to lazily read signals.
    """
    try:
        # not released in xarray 2024.7.0, this will be released soon
        from xarray import DataTree
    except:
        # this need the experimental DataTree in pypi xarray-datatree
        try:
            from datatree import DataTree
        except:
            raise ImportError("use xarray dev branch or pip install xarray-datatree")

    signal_buffers = rawio_reader.header["signal_buffers"]
    buffer_ids = signal_buffers["id"]

    tree = DataTree(name="root")

    num_block = rawio_reader.header["nb_block"]
    for block_index in range(num_block):
        block = DataTree(name=f"block{block_index}", parent=tree)
        num_seg = rawio_reader.header["nb_segment"][block_index]
        for seg_index in range(num_seg):
            segment = DataTree(name=f"segment{seg_index}", parent=block)
            for buffer_id in buffer_ids:
                ds = to_xarray_dataset(rawio_reader, block_index=block_index, seg_index=seg_index, buffer_id=buffer_id)
                DataTree(data=ds, name=ds.attrs["name"], parent=segment)

    return tree
